--[[
	### ItemValue ###
	ItemValue StatSetFormula Class
	
--]]

-- Libraries
local L = returnkey -- AceLibrary("AceLocale-2.2"):new("ItemValue")
local AceOO = AceLibrary("AceOO-2.0")
local ItemBonusLib = AceLibrary("ItemBonusLib-1.0")

local db

local Formula = AceOO.Class()

function Formula.LoadSets()
	db = ItemValue:AcquireDBNamespace("Formula")
	
	for SetName in pairs(db.profile) do
		ItemValue.RegisterSet(Formula:new(SetName))
	end
end

function Formula.prototype:init(Name, Enabled, argFormula)
	Formula.super.prototype.init(self)
	
	if not db.profile[Name] then
		db.profile[Name] = {
			Enabled = (Enabled == nil) or Enabled,
			Formula = argFormula or "",
		}
	end
	
	self.Name = Name
	self.Enabled = db.profile[Name].Enabled
	self.Formula = db.profile[Name].Formula
	
	self.Options = {
		type = "group",
		name = Name,
		desc = string.format(L["Settings for Set \"%s\"."], Name),
		args = {
			name = {
				order = 1,
				type  = "text",
				name  = L["Name"],
				desc  = L["Renames the Set."],
				set   = function(v) ItemValue.RenameSet(self, v) end,
				get   = function() return self.Name end,
				usage = "<New Name>",
				vaidate = tostring,
			},
			enabled = {
				order = 2,
				type  = "toggle",
				name  = L["Enabled"],
				desc  = L["Defines if the Item Value for this set should be calculated."],
				set   = function(v)
					self.Enabled = v
					if not v then v = nil end
					db.profile[name].Enabled = v
					ItemValue.ClearCache()
				end,
				get   = function() return self.Enabled end,
			},
			delete = {
				type = "execute",
				order = 3,
				name = "Delete",
				desc = "Delets this Set.",
				func = ItemValue.UnregisterSet,
				passValue = self,
			},
			formula = {
				order = 100,
				type  = "text",
				name  = L["Formula"],
				get   = function() return self.Formula or "" end,
				set   = function(v)
					self.Formula = v
					db.profile[Name].Formula = v
					ItemValue:ClearCache()
				end
			},
		},
	}

end

function Formula.prototype:tostring()
	return "ItemValue Formula " .. self.Name
end

function Formula.prototype:OnDelete()
	db.profile[self.Name] = nil
end

function Formula.prototype:GetEquipValue(eq)
	Formula.Data = setmetatable(eq, {["__index"] = function() return 0 end})
	local Points = loadstring("local math=math;setfenv(1, ItemValue.classes.Formula.Data);return " .. self.Formula)()
	Formula.Data = nil
	
	return Points or 0
end

function Formula.prototype:GetItemValue(ItemString)
	return self:GetEquipValue(ItemBonusLib:ScanItem(ItemString, true))
end

function Formula.prototype:Serialize()
	return {Type = "Formula", Name = self.name, Enabled = self.enabled, Formula = self.Formula}
end

function Formula:Deserialize(t)
	return Formula:new(t.Name, t.Enabled, t.Formula)
end

ItemValue.RegisterSetType("Formula", Formula)
