--[[
	### ItemValue ###
	ItemValue StatSet Class
	
--]]

-- Libraries
local L = returnkey -- AceLibrary("AceLocale-2.2"):new("ItemValue")
local AceOO = AceLibrary("AceOO-2.0")
local ItemBonusLib = AceLibrary("ItemBonusLib-1.0")

local db

local ItemBonusLibStats = {
	{"STR", "AGI", "STA", "INT", "SPI", "HEALTH", "HEALTHREG"},
	{"ATTACKPOWER", "ATTACKPOWERUNDEAD", "ATTACKPOWERFERAL", "CR_CRIT", "CR_HIT", "CR_HASTE", "RANGEDATTACKPOWER", "RANGEDCRIT"},
	{"DMG", "DMGUNDEAD", "ARCANEDMG", "FIREDMG", "FROSTDMG", "HOLYDMG", "NATUREDMG", "SHADOWDMG", "CR_SPELLCRIT", "CR_SPELLHIT", "SPELLPEN", "CR_SPELLHASTE"},
	{"HEAL", "MANAREG", "MANA"},
	{"ARMOR", "CR_RESILIENCE", "CR_DEFENSE", "CR_BLOCK", "BLOCKVALUE", "CR_DODGE", "CR_PARRY"},
	{"CR_WEAPON", "CR_WEAPON_AXE", "CR_WEAPON_DAGGER", "CR_WEAPON_MACE", "CR_WEAPON_SWORD", "CR_WEAPON_SWORD_2H"},
	{"ARCANERES", "FIRERES", "NATURERES", "FROSTRES", "SHADOWRES"},
}

local ItemBonusLibHeaders = {
	"Base Stats",
	"Physical DPS",
	"Magical DPS",
	"Healing",
	"Tanking",
	"Weapon Skill",
	"Resistances"
}

local StatSet = AceOO.Class()

function StatSet.LoadSets()
	db = ItemValue:AcquireDBNamespace("StatSet")
	
	for SetName in pairs(db.profile) do
		ItemValue.RegisterSet(StatSet:new(SetName))
	end
end

function StatSet.prototype:init(Name, Enabled, Stats)
	StatSet.super.prototype.init(self)
	
	if not db.profile[Name] then
		db.profile[Name] = {
			Enabled = (Enabled == nil) or Enabled,
			Stats = Stats or {},
		}
	end
	
	self.Name = Name
	self.Enabled = db.profile[Name].Enabled
	self.Stats = db.profile[Name].Stats
	
	self.Options = {
		type = "group",
		name = Name,
		desc = string.format(L["Settings for Set \"%s\"."], Name),
		args = {
			name = {
				order = 1,
				type  = "text",
				name  = L["Name"],
				desc  = L["Renames the Set."],
				set   = function(v) ItemValue.RenameSet(self, v) end,
				get   = function() return self.Name end,
				usage = "<New Name>",
				vaidate = tostring,
			},
			enabled = {
				order = 2,
				type  = "toggle",
				name  = L["Enabled"],
				desc  = L["Defines if the Item Value for this set should be calculated."],
				set   = function(v)
					self.Enabled = v
					if not v then v = nil end
					db.profile[Name].Enabled = v
					ItemValue.ClearCache()
				end,
				get   = function() return self.Enabled end,
			},
			delete = {
				type = "execute",
				order = 3,
				name = "Delete",
				desc = "Delets this Set.",
				func = ItemValue.UnregisterSet,
				passValue = self,
			},
			statvalues = {
				order = 100,
				type  = "header",
				name  = L["Stat Values"],
			},
		},
	}
	
	local n = 100
	
	local get = function(k)
		return self.Stats[k] or 0
	end
	local set = function(k, v)
		v = tonumber(v)
		if v == 0 then v = nil end
		self.Stats[k] = v
		ItemValue:ClearCache() 
	end
	
	for h, Stats in pairs(ItemBonusLibStats) do
		n = n + 1
		self.Options.args["header" .. n] = {
			order = n,
			type = "header",
			name = L[ItemBonusLibHeaders[h]],
		}
		for _, iblname in pairs(Stats) do
			n = n + 1
			self.Options.args[iblname] = {
				order	= n,
				type	= "text",
				name	= ItemBonusLib:GetBonusFriendlyName(iblname),
				desc	= ItemBonusLib:GetBonusFriendlyName(iblname),
				get		= get,
				set		= set,
				validate = tonumber,
				usage	= "<Number>",
				passValue = iblname,
			}
		end
	end
end

function StatSet.prototype:tostring()
	return "ItemValue StatSet " .. self.Name
end

function StatSet.prototype:OnDelete()
	db.profile[self.Name] = nil
end

function StatSet.prototype:GetEquipValue(eq)
	local Points = 0
	
	if eq then
		for StatName, StatValue in pairs(self.Stats) do
			if eq[StatName] and StatValue then
				Points = Points + (eq[StatName] * StatValue)
			end
		end
	end
	
	return Points
end

function StatSet.prototype:GetItemValue(ItemString)
	return self:GetEquipValue(ItemBonusLib:ScanItem(ItemString, true))
end

function StatSet.prototype:Serialize()
	return {Type = "StatSet", Name = self.name, Enabled = self.Enabled, Stats = self.Stats}
end

function StatSet:Deserialize(t)
	return StatSet:new(t.Name, t.Enabled, t.Stats)
end

ItemValue.RegisterSetType("StatSet", StatSet)
