--[[
	### ItemValue ###
	ItemValue Core
	
	Used Libraries:
	Ace2, GratuityLib, Deformat, ItemBonusLib, TipHookerLib, Waterfall-1.0
--]]

ItemValue = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceHook-2.1", "AceDB-2.0", "AceDebug-2.0", "AceComm-2.0")
ItemValue.revision = tonumber(string.sub("$Revision: 40766 $", 12, -3))
ItemValue.version = "r" .. (ItemValue.revision or 0)
ItemValue.date = string.sub("$Date: 2007-06-19 18:20:46 -0400 (Tue, 19 Jun 2007) $", 8, 17)

-- Libraries
local L = returnkey -- AceLibrary("AceLocale-2.2"):new("ItemValue")
local AceOO = AceLibrary("AceOO-2.0")
local TipHooker = AceLibrary("TipHooker-1.0")
local Gratuity = AceLibrary("Gratuity-2.0")
local Waterfall = AceLibrary("Waterfall-1.0")
local ItemBonusLib = AceLibrary("ItemBonusLib-1.0")
local BabbleClass = AceLibrary("Babble-Class-2.2")

local string = string
local table = table
local setmetatable = setmetatable
local pairs = pairs
local type = type
local unpack = unpack
local select = select
local math = math
local floor = floor
local ceil = ceil
local strsplit = strsplit
local tostring = tostring
local tonumber = tonumber
local assert = assert
local loadstring = loadstring

local CreateFrame = CreateFrame
local GetItemInfo = GetItemInfo
local IsEquippableItem = IsEquippableItem
local GetMouseFocus = GetMouseFocus
local GetInventoryItemLink = GetInventoryItemLink
local IsEquippedItem = IsEquippedItem
local GetItemGem = GetItemGem

local GameFontWhite = GameFontWhite
local CharacterModelFrame = CharacterModelFrame

local UnitClass = UnitClass
local UnitLevel = UnitLevel
local UnitName = UnitName
local UnitExists = UnitExists
local UnitPlayerControlled = UnitPlayerControlled
local UnitIsFriend = UnitIsFriend
local CheckInteractDistance = CheckInteractDistance

local self = ItemValue
setfenv(1, ItemValue)

-- Table for SetTypes
classes = {}

-- AceOptions Table
local Options = {
	type = "group",
	args = {
		header = {
			order = 1,
			type  = "header",
			name  = "|cFF77BBFF".."ItemValue".." |c88888888"..version,
			icon  = "Interface\\Icons\\Ability_Warrior_Disarm",
			iconHeight = 24,
			iconWidth = 24,
		},
		GUI = {
			order 	= 2,
			type	= "execute",
			name	= L["Waterfall GUI"],
			desc	= L["Opens the ItemValue configuration GUI."],
			func	= function() Waterfall:Open("ItemValue") end,
			wfHidden = true,
		},
		bj = {
			type	= "text",
			name	= L["Best Jewels"],
			desc	= L["Calculates the best jewel combination for a given Item Link."],
			get		= false,
			set		= function(ItemLink)
				for SetName, Set in pairs(SetByName) do
					if Set.Enabled then
						PrintBestJewels(self, SetName, ItemLink)
					end
				end
			end,
			usage	= "<Item Link>",
			validate = function(ItemLink) return GetItemInfo(ItemLink) end,
			wfHidden = true,
		},
		SendSet = {
			order = 1,
			type  = "text",
			name = "Send Set",
			desc = "Sends a set to another ItemValue user.",
			usage = "<Player Name> <Set Name>",
			get = false,
			set = "SendStatSet",
			validate = function(PlayerName, SetName) return (PlayerName and SetByName[SetName]) end,
			input = true,
		},
		Sets = {
			order	= 3,
			type	= "group",
			name	= L["Sets"],
			desc	= L["Configuration of ItemValue Sets."],
			args	= {
				New = {
					type = "group",
					name = L["New"],
					desc = L["Creates new Sets."],
					args = {},
				},
			},
		},
		Settings = {
			type = "group",
			name = L["Settings"],
			desc = L["General settings about ItemValue."],
			args = {
				Modifications = {
					order	= 1,
					type	= "group",
					name	= L["Item Modifications"],
					desc	= L["How enchants and jewels should be handled."],
					args	= {
						Enchants = {
							order	= 1,
							type	= "group",
							name	= "Enchants",
							desc	= "Enchant related settings.",
							args	= {
								Enchants = {
									order	= 1,
									type	= "text",
									name	= L["Enchants"],
									desc	= L["Defines if item enchants should be included in the calculation."],
									get		= function() return db.profile.Modifications.Enchants end,
									set		= function(v) db.profile.Modifications.Enchants = v ClearCache() end,
									validate = {L["Ignore"], L["Do nothing"], L["Always own"], L["Fill with own"]},
								},
							},
						},
						Jewels = {
							order	= 2,
							type	= "group",
							name	= "Jewels",
							desc	= "Jewel related settings.",
							args	= {
								Jewels = {
									order	= 1,
									type	= "text",
									name	= L["Jewels"],
									desc	= L["Defines if jewels should be included in the calculation."],
									get		= function() return db.profile.Modifications.Jewels end,
									set		= function(v) db.profile.Modifications.Jewels = v ClearCache() end,
									validate = {L["Ignore"], L["Do nothing"], L["Always default"], L["Fill with default"], L["Best"]},
								},
								DefaultJewel = {
									order	= 2,
									type	= "text",
									name	= L["Default Jewel"],
									desc	= L["What jewel should be used for \"Always Default\" and \"Fill with Default\" modes."],
									disabled = function()
										return db.profile.Modifications.MultipleDefaultJewels or 
										not((db.profile.Modifications.Jewels == L["Always default"] or 
										db.profile.Modifications.Jewels == L["Fill with default"]))
									end,
									get		= function() return db.profile.Modifications.DefaultJewel end,
									set		= function(v) db.profile.Modifications.DefaultJewel = v ClearCache() end,
									validate = function(v)
										for JewelName in pairs(JewelByName) do
											if v == JewelName then return true end
										end
										return
									end,
									usage = L["<Jewel Name>"],
								},
								MultipleDefaultJewels = {
									order	= 3,
									type	= "toggle",
									name	= L["Multiple Default Jewels"],
									desc	= L["Allows you to define different default jewels for diffently colorored sockets."],
									get		= function() return db.profile.Modifications.MultipleDefaultJewels end,
									set		= function(v) db.profile.Modifications.MultipleDefaultJewels = v ClearCache() end,
									disabled = function()
										return not(db.profile.Modifications.Jewels == L["Always default"] or 
										db.profile.Modifications.Jewels == L["Fill with default"])
									end,
								},
								DefaultJewelRed = {
									order	= 4,
									type	= "text",
									name	= L["Default Jewel: Red"],
									desc	= L["What jewel should be used for \"Always Default\" and \"Fill with Default\" modes."],
									disabled = function()
										return not((db.profile.Modifications.Jewels == L["Always default"] or 
										db.profile.Modifications.Jewels == L["Fill with default"]) and 
										db.profile.Modifications.MultipleDefaultJewels)
									end,
									get		= function() return db.profile.Modifications.DefaultJewels.Red end,
									set		= function(v) db.profile.Modifications.DefaultJewels.Red = v ClearCache() end,
									validate = function(v)
										for JewelName in pairs(JewelByName) do
											if v == JewelName then return true end
										end
										return
									end,
									usage = L["<Jewel Name>"],
								},
								DefaultJewelYellow = {
									order	= 5,
									type	= "text",
									name	= L["Default Jewel: Yellow"],
									desc	= L["What jewel should be used for \"Always Default\" and \"Fill with Default\" modes."],
									disabled = function()
										return not((db.profile.Modifications.Jewels == L["Always default"] or 
										db.profile.Modifications.Jewels == L["Fill with default"]) and 
										db.profile.Modifications.MultipleDefaultJewels)
									end,
									get		= function() return db.profile.Modifications.DefaultJewels.Yellow end,
									set		= function(v) db.profile.Modifications.DefaultJewels.Yellow = v ClearCache() end,
									validate = function(v)
										for JewelName in pairs(JewelByName) do
											if v == JewelName then return true end
										end
										return
									end,
									usage = L["<Jewel Name>"],
								},
								DefaultJewelBlue = {
									order	= 6,
									type	= "text",
									name	= L["Default Jewel: Blue"],
									desc	= L["What red jewel should be used for \"Always Default\" and \"Fill with Default\" modes."],
									disabled = function()
										return not((db.profile.Modifications.Jewels == L["Always default"] or 
										db.profile.Modifications.Jewels == L["Fill with default"]) and 
										db.profile.Modifications.MultipleDefaultJewels)
									end,
									get		= function() return db.profile.Modifications.DefaultJewels.Blue end,
									set		= function(v) db.profile.Modifications.DefaultJewels.Blue = v ClearCache() end,
									validate = function(v)
										for JewelName in pairs(JewelByName) do
											if v == JewelName then return true end
										end
										return
									end,
									usage = L["<Jewel Name>"],
								},
								DefaultJewelMeta = {
									order	= 7,
									type	= "text",
									name	= L["Default Jewel: Meta"],
									desc	= L["What red jewel should be used for \"Always Default\" and \"Fill with Default\" modes."],
									disabled = function()
										return not((db.profile.Modifications.Jewels == L["Always default"] or 
										db.profile.Modifications.Jewels == L["Fill with default"]) and 
										db.profile.Modifications.MultipleDefaultJewels)
									end,
									get		= function() return db.profile.Modifications.DefaultJewels.Meta end,
									set		= function(v) db.profile.Modifications.DefaultJewels.Meta = v ClearCache() end,
									validate = function(v)
										for JewelName in pairs(JewelByName) do
											if v == JewelName then return true end
										end
										return
									end,
									usage = L["<Jewel Name>"],
								},
							},
						},
					},
				},
				Tooltip = {
					order	= 2,
					type	= "group",
					name	= L["Tooltip"],
					desc	= L["Tooltip Display Settings"],
					args	= {
						enable = {
							order	= 1,
							type	= "toggle",
							name	= L["Enable"],
							desc	= L["Enables ALL tooltip display."],
							get		= function() return db.profile.Tooltip.Enable end,
							set		= function(v)
								db.profile.Tooltip.Enable = v
								if v then
									TipHooker:Hook(AddDataToTooltip, "item")
								else
									TipHooker:Unhook(AddDataToTooltip, "item")
								end
								ClearCache()
							end,
						},
						showPoints = {
							order	= 2,
							type	= "toggle",
							name	= L["Show Points"],
							desc	= L["Shows the Setpoints of an item in its tooltip."],
							get		= function() return db.profile.Tooltip.ShowPoints end,
							set		= function(v) ClearCache() db.profile.Tooltip.ShowPoints = v end,
						},
						compare = {
							order	= 3,
							type	= "text",
							name	= L["Compare"],
							desc	= L["If and how informations of other equipped items should be displayed."],
							get		= function() return db.profile.Tooltip.Compare end,
							set		= function(v) ClearCache() db.profile.Tooltip.Compare = v end,
							validate = {L["Don't Compare"], L["Absolute"], L["Delta"], L["Percent"]},
						},
						swapcolor = {
							order	= 4,
							type	= "toggle",
							name	= L["Swap Colors"],
							desc	= L["Swap the comparison colors?"],
							get		= function() return db.profile.Tooltip.SwapColors end,
							set		= function(v) ClearCache() db.profile.Tooltip.SwapColors = v end,
						},
						swapcomparison = {
							order	= 4,
							type	= "toggle",
							name	= L["Swap Comparison"],
							desc	= L["Swap the comparison (+5 gets -5)?"],
							get		= function() return db.profile.Tooltip.SwapComparison end,
							set		= function(v) ClearCache() db.profile.Tooltip.SwapComparison = v end,
						},
						rightside = {
							order	= 5,
							type	= "toggle",
							name	= L["Right Side"],
							desc	= L["Displays item values on the right side."],
							get		= function() return db.profile.Tooltip.RightSide end,
							set		= function(v) ClearCache() db.profile.Tooltip.RightSide = v end,
						},
					},
				},
				EQValue = {
					order = 3,
					type = "group",
					name = L["Player Equip Value"],
					desc = L["All ItemValues of a player"],
					args = {
						self = {
							order = 1,
							type = "group",
							name = "Self",
							desc = "Own equipment value",
							args = {
								enableself = {
									order	= 2,
									type	= "toggle",
									name	= L["Enable"],
									desc	= L["Shows own equipment value in the charsheet."],
									get		= function() return db.profile.PlayerValue.Self end,
									set		= function(v)
										db.profile.PlayerValue.Self = v
										if v then
											if not PointsFrame then PointsFrame = CreatePlayerValueFrame() end
											PointsFrame:Show()
										else
											if PointsFrame then PointsFrame:Hide() end
										end
										--if db.profile.PlayerValue.Self or db.profile.PlayerValue.Inspect then
										--	RegisterEvent(self, "ItemBonusLib_Update", UpdatePlayerEquipment)
										--else
										--	UnregisterEvent(self, "ItemBonusLib_Update")
										--end
										ClearCache()
									end,
								},
							}
						},
						inspect = {
							order = 2,
							type = "group",
							name = "Inspect",
							desc = "Inspect equipment value",
							args = {
								EnableInspect = {
									order	= 1,
									type	= "toggle",
									name	= L["Enable"],
									desc	= L["Print out targets Equipment Value while inspecting."],
									get		= function() return db.profile.PlayerValue.Inspect end,
									set		= function(v)
										db.profile.PlayerValue.Inspect = v
										if v then
											Hook(self, "InspectUnit", true)
										else
											Unhook(self, "InspectUnit")
										end
										--if db.profile.PlayerValue.Self or db.profile.PlayerValue.Inspect and not IsEventRegistered("ItemBonusLib_Update") then
										--	RegisterEvent(self, "ItemBonusLib_Update", UpdatePlayerEquipment)
										--else
										--	UnregisterEvent(self, "ItemBonusLib_Update")
										--end
									end,
								},
								ShowPoints = {
									order	= 2,
									type	= "toggle",
									name	= L["Show Points"],
									desc	= L["Shows the equipment value of a player."],
									get		= function() return db.profile.PlayerValue.ShowPoints end,
									set		= function(v) db.profile.PlayerValue.ShowPoints = v end,
								},
								Compare = {
									order	= 3,
									type	= "text",
									name	= L["Compare Mode"],
									desc	= L["How your equipment points should be compared to your targets."],
									get		= function() return db.profile.PlayerValue.Compare end,
									set		= function(v) db.profile.PlayerValue.Compare = v end,
									validate = {L["Don't Compare"], L["Absolute"], L["Delta"], L["Percent"]},
								},
								SwapColor = {
									order	= 4,
									type	= "toggle",
									name	= L["Swap Colors"],
									desc	= L["Swap the comparison colors?"],
									get		= function() return db.profile.PlayerValue.SwapColors end,
									set		= function(v) ClearCache() db.profile.PlayerValue.SwapColors = v end,
								},
								SwapComparison = {
									order	= 5,
									type	= "toggle",
									name	= L["Swap Comparison"],
									desc	= L["Swap the comparison (+5 becomes -5)?"],
									get		= function() return db.profile.PlayerValue.SwapComparison end,
									set		= function(v) ClearCache() db.profile.PlayerValue.SwapComparison = v end,
								},
							},
						},
					},
				},
			},
		},
		Tools = {
			type = "group",
			name = L["Tools"],
			desc = L["Jewel Tools"],
			args = {
				BestJewels = {
					order	= 7,
					type	= "text",
					name	= L["Calculate Best Jewels"],
					desc	= L["Calculates the best possible Jewel combination for an ItemLink."],
					usage 	= L["<Set Name> <ItemLink>"],
					get		= false,
					set		= "PrintBestJewels",
					input	= true,
				},
				ListJewels = {
					order	= 9,
					type	= "text",
					name	= L["List Jewels"],
					desc	= L["Lists all known Jewels."],
					get		= false,
					usage	= "\"all\" OR \"enabled\" OR \"disabled\" OR <Part of Jewelname>",
					set	= function(v)
						local color = {"|cFFFFFFFF[", "|cFF1EFF00[", "|cFF0070DD[", "|cFFA335EE[", "|cFFFF8000["}
						Print(self, string.format(L["Printing \"%s\" jewels..."], v))
						local shown = 0
						local total = 0
						for Name, Jewel in pairs(JewelByName) do
							if (v == "all") or
							 (v == "enabled" and db.profile.EnabledJewels[Name]) or
							 (v == "disabled" and not db.profile.EnabledJewels[Name]) or
							 (string.find(string.lower(Name), string.lower(v))) then
								
								local _, ItemLink = GetItemInfo(Jewel[2])
								if not ItemLink then
									-- Not in local cache -> get color data from db
									ItemLink = color[Jewel.Quality] .. Name .. "]|r (*)"
								end
								Print(self, " ", ItemLink)
								shown = shown + 1
							end
							total = total + 1
						end
						Print(self, string.format(L["%s/%s Jewels shown."], shown, total))
					end,
				},
			},
		},
	},
}

local DefaultSettings = {
	Tooltip = {
		Enable = true,
		RightSide = false,
		ShowPoints = true,
		Compare = L["Absolute"],
		SwapColors = false,
		SwapComparison = false,
	},
	Modifications = {
		Enchants = L["Always own"],
		Jewels = L["Fill with default"],
		DefaultJewel = "",
		MultipleDefaultJewels = false,
		DefaultJewels = {
			Meta = "",
			Red = "",
			Yellow = "",
			Blue = "",
		},
	},
	PlayerValue = {
		Self = false,
		Inspect = false,
		Compare = L["Absolute"],
		SwapColors = false,
		SwapComparison = false,
	},
	EnabledJewels = {
		["*"] = true,
	},
	Jewels = {},
}

function OnInitialize(self)
	RegisterDB(self, "ItemValueDB", "ItemValueDBPerChar")
	RegisterDefaults(self, "profile", DefaultSettings)
	SetCommPrefix(self, "ItemValue")
	RegisterComm(self, "ItemValue", "WHISPER")
	
	SocketAmountCache = setmetatable({}, {__mode = "kv"})
	SocketInfoCache = setmetatable({}, {__mode = "kv"})
	TooltipTextCache = setmetatable({}, {__mode = "kv"})
	TooltipTextCache2 = setmetatable({}, {__mode = "kv"})
	
	RegisterChatCommand(self, "/iv", Options)
	--Waterfall:Register("ItemValue", "aceOptions", Options, "hideTreeRoot", true, "colorR", 0.2, "colorG", 0.5, "colorB", 0.8)
	Waterfall:Register("ItemValue", "aceOptions", Options, "treeType", "SECTIONS", "colorR", 0.2, "colorG", 0.5, "colorB", 0.8)
end

function OnEnable(self, first)
	if first then
		JewelByName, JewelByEnchantId = LoadJewels()
		Options.args.Settings.args.Jewels = LoadJewelFilters()
		
		-- Table for Sets
		SetByName = {}
		
		-- Load sets
		for _, SetClass in pairs(classes) do
			SetClass.LoadSets()
		end
		
		ClearCache()
	end
	
	if db.profile.Tooltip.Enable then
		TipHooker:Hook(AddDataToTooltip, "item")
	end
	
	if db.profile.PlayerValue.Self and not PlayerValueFrame then
		PlayerValueFrame = CreatePlayerValueFrame()
	end
	
	if db.profile.PlayerValue.Inspect then
		Hook(self, "InspectUnit", true)
	end
	
	--if db.profile.PlayerValue.Self or db.profile.PlayerValue.Inspect then
	RegisterEvent(self, "ItemBonusLib_Update", UpdatePlayerEquipment)
	--end
end

function OnDisable()
	TipHooker:Unhook(AddValueToTooltip, "item")
	if PlayerValueFrame then PlayerValueFrame:Hide() end
end

function ClearCache()
	for key in pairs(TooltipTextCache) do
		TooltipTextCache[key] = nil
	end
	for key in pairs(TooltipTextCache2) do
		TooltipTextCache2[key] = nil
	end
	
	UpdatePlayerValueFrame()
end

-- Self Equipment Value Frame
function CreatePlayerValueFrame()
	local frame = CreateFrame("Frame", nil, CharacterModelFrame)
	frame:SetFrameStrata("HIGH")
	frame:SetWidth(256)
	frame:SetHeight(128)
	
	frame.text = frame:CreateFontString(nil, "HIGH")
	frame.text:SetFontObject(GameFontWhite)
	frame.text:SetJustifyH("LEFT") 
	frame.text:SetJustifyV("BOTTOM")
	frame.fontstring = frame.text
	
	frame.text:SetText("")
	frame.text:SetAllPoints(frame)
	
	frame:SetPoint("BOTTOMLEFT", 2, 24)
	frame:Show()
	
	return frame
end

function UpdatePlayerEquipment()
	local eq = ItemBonusLib:GetUnitEquipment("player")
	local details = ItemBonusLib:BuildBonusSet(eq)
	EquipStats = ItemBonusLib:MergeDetails(details)
	
	ClearCache()
end

function UpdatePlayerValueFrame()
	if not PlayerValueFrame then return end
	local PlayerValueString = {}
	
	for SetName, Set in pairs(SetByName) do
		if Set.Enabled then
			local points = floor(Set:GetEquipValue(EquipStats))
			table.insert(PlayerValueString, SetName .. ": " .. points)
		end
	end
	PlayerValueString = table.concat(PlayerValueString, "\n")
	
	PlayerValueFrame.text:SetText(PlayerValueString)
end

function InspectUnit(self, ...)
	if (UnitExists("target") and UnitPlayerControlled("target") and
	 UnitIsFriend("target", "player") and CheckInteractDistance("target", 1)) then
	 
		Print(self, "Inspecting", UnitLevel("target"), "|cFF" .. BabbleClass:GetHexColor(UnitClass("target")) .. (UnitName("target")) .. "|r...")
		
		local teq = ItemBonusLib:GetUnitEquipment("target")
		local tdetails = ItemBonusLib:BuildBonusSet(teq)
		local tbonuses = ItemBonusLib:MergeDetails(tdetails)
		
		for SetName, Set in pairs(SetByName) do
			if Set.Enabled then
				local points = floor(Set:GetEquipValue(EquipStats))
				local tpoints = floor(Set:GetEquipValue(tbonuses))
				if tpoints then
					Print(self, " " .. SetName .. ": " .. Compare(self.db.profile.PlayerValue.Compare, self.db.profile.PlayerValue.SwapColors, self.db.profile.PlayerValue.SwapComparison, self.db.profile.PlayerValue.ShowPoints, tpoints, points))
				end
			end
		end
	end
	
	self.hooks["InspectUnit"](...)
end

-- Set Handling
function RegisterSet(Set)
	local SetName = Set.Name
	if SetByName[SetName] then
		Print(self, string.format(L["Error: Set \"%s\" already exists!"], SetName))
		return
	end
	
	Options.args.Sets.args[SetName] = Set.Options
	SetByName[SetName] = Set
	ClearCache()
	TriggerEvent(self, "ItemValue_SetsChanged")
end

function UnregisterSet(Set)
	local open = Waterfall:IsOpen("ItemValue")
	if open then Waterfall:Close("ItemValue") end
	
	local SetName = Set.Name
	if not SetByName[SetName] then
		return string.format(L["Error: Set \"%s\" doesn't exist!"], SetName)
	end
	
	Set:OnDelete()
	Options.args.Sets.args[SetName] = nil
	SetByName[SetName] = nil
	ClearCache()
	TriggerEvent(self, "ItemValue_SetsChanged")
	if open then Waterfall:Open("ItemValue") end
end

function RenameSet(OldSet, NewName)
	-- Serializes old set, changes the name and deserializes to a new object
	assert(OldSet)
	assert(not SetByName[NewName])
	
	local t = OldSet:Serialize()
	t.Name = NewName
	local NewSet = classes[t.Type]:Deserialize(t)
	assert(NewSet)
	
	RegisterSet(NewSet)
	UnregisterSet(OldSet)
	TriggerEvent(self, "ItemValue_SetsChanged")
end

function RegisterSetType(SetClassName, SetClass)
	assert(SetClass)
	assert(SetClassName)
	
	classes[SetClassName] = SetClass
	Options.args.Sets.args.New.args[SetClassName] = {
		type	= "text",
		name	= string.format(L["New %s"], SetClassName),
		desc	= string.format(L["Creates a new %s."], SetClassName),
		usage	= L["<Name>"],
		get		= false,
		set		= function(name)
			Print(self, string.format(L["Creating new %s \"%s\"..."], SetClassName, name))
			RegisterSet(SetClass:new(name))
			ClearCache()
		end,
		validate = function(v)
			return v and type(v) == "string" and string.len(v) > 0
		end,
	}
	
	if SetByName then
		SetClass.LoadSets()
	end
end

function SendStatSet(self, PlayerName, SetName)
	Print(self, string.format(L["Sending Set \"%s\" to player \"%s\"..."], SetName, PlayerName))
	SendCommMessage(self, "WHISPER", PlayerName, SetByName[SetName]:Serialize())
end

function OnCommReceive(self, prefix, sender, distribution, t)
	assert(t and type(t) == "table", "Set invalid.")
	assert(t.Name and type(t.Name) == "string", "Name invalid.")
	assert(t.Type and type(t.Type) == "string", "Type invalid.")
	assert(classes[t.Type], t.Type .. " Type unknown.")
	
	local SetType = t.Type
	Print(self, string.format(L["Received \"%s\" named \"%s\" from player \"%s\"."], SetType, t.Name, sender))
	t.Name = t.Name .. " - " .. sender
	
	local Set = classes[SetType]:Deserialize(t)
	if not Set then Print(self, string.format(L["ERROR: Couldn't Deserialize \"%s\"!"], SetType)) return end
	RegisterSet(Set)
end

function AddDataToTooltip(Tooltip, _, argItemLink, argItemString)
	local ItemString = argItemString or ItemLinkToItemString(argItemLink) or ItemLinkToItemString(Tooltip:GetItemLink())
	
	local text = TooltipTextCache[ItemString]
	
	if db.profile.Tooltip.RightSide then
		local text2 = TooltipTextCache2[ItemString]
		if not (text and text2) then
			text, text2 = GetTooltipText(ItemString)
			TooltipTextCache[ItemString] = text
			TooltipTextCache2[ItemString] = text2
		end
		Tooltip:AddDoubleLine(text, text2)
	else
		if not text then
			text = GetTooltipText(ItemString)
			TooltipTextCache[ItemString] = text
		end
		Tooltip:AddLine(text)
	end
	
	Tooltip:Show()
end

function GetTooltipText(ItemString)
	local slot1, slot2
	local ItemLink1, ItemLink2, ItemLink3
	local ItemString1, ItemString2, ItemString3 = ItemString
	
	local LinkType, iId1, eId1, jId11, jId21, jId31, jId41, sId1, uId1 = strsplit(":", (ItemString1))
	if LinkType ~= "item" then return "", "" end
	
	local _, _, _, _, _, _, ItemType = GetItemInfo(ItemString1)
	if not (IsEquippableItem(ItemString1) or ItemType == L["Gem"]) then return "", "" end
	
	local text, text2
	
	local iId2, eId2, jId12, jId22, jId32, jId42, sId2, uId2
	local iId3, eId3, jId13, jId23, jId33, jId43, sId3, uId3
	
	-- Unknown Jewel?
	if not JewelByEnchantId[tonumber(jId11)] then RegisterNewJewel(jId11) end
	if not JewelByEnchantId[tonumber(jId21)] then RegisterNewJewel(jId21) end
	if not JewelByEnchantId[tonumber(jId31)] then RegisterNewJewel(jId31) end
	
	local _, _, _, _, _, _, _, _, itemEquipLoc = GetItemInfo(ItemString1)
	local frame = GetMouseFocus() and GetMouseFocus():GetName() or ""
	local ownItemFrame = (frame == "WorldFrame" or string.find(frame,"^Character.*Slot$") or string.find(frame,"^TempEnchant%d+$"))
	local ownItem = IsEquippedItem(ItemString1)
	
	if itemEquipLoc and not(db.profile.Tooltip.Compare == "None" or ownItemFrame or ownItem) then
		slot1, slot2 = GetInventorySlot(ItemString1)
		
		if slot1 then
			ItemString2 = ItemLinkToItemString(GetInventoryItemLink("player", slot1))
			if ItemString2 then
				_, iId2, eId2, jId12, jId22, jId32, jId42, sId2, uId2 = strsplit(":", ItemString2)
			end
		end
		if slot2 then
			ItemString3 = ItemLinkToItemString(GetInventoryItemLink("player", slot2))
			if ItemString3 then
				_, iId3, eId3, jId13, jId23, jId33, jId43, sId3, uId3 = strsplit(":", ItemString3)
			end
		end
	end
	
	-- Strip enchant and jewel infomations from item string
	if db.profile.Modifications.Enchants == L["Always own"] then
		-- Set item enchant to enchant from equipped item 
		if eId2 or eId3 then
			eId1 = eId2 or eId3
		end
	elseif db.profile.Modifications.Enchants == L["Fill with own"] then
		if eId1 == "0" and (eId2 or eId3) then
			eId1 = eId2 or eId3
		end
	elseif db.profile.Modifications.Enchants == L["Ignore"] then
		eId1 = "0"
		if iId2 then eId2 = "0" end
		if iId2 then eId3 = "0" end
	end
	
	local SocketCount1, SocketCount2, SocketCount3 = GetSocketCount(iId1)
	if iId2 then SocketCount2 = GetSocketCount(iId2) end
	if iId3 then SocketCount3 = GetSocketCount(iId3) end
	
	if SocketCount1 > 0 or (SocketCount2 and SocketCount2 > 0) or (SocketCount3 and SocketCount3 > 0) then
		local DefaultJewel = JewelByName[db.profile.Modifications.DefaultJewel] or EmptySocket
		if db.profile.Modifications.Jewels == L["Fill with default"] then
			jId11 = (SocketCount1 > 0 and jId11 == "0" and DefaultJewel[1]) or jId11
			jId21 = (SocketCount1 > 1 and jId21 == "0" and DefaultJewel[1]) or jId21
			jId31 = (SocketCount1 > 2 and jId31 == "0" and DefaultJewel[1]) or jId31
			if iId2 then
				jId12 = (SocketCount2 > 0 and jId12 == "0" and DefaultJewel[1]) or jId12
				jId22 = (SocketCount2 > 1 and jId22 == "0" and DefaultJewel[1]) or jId22
				jId32 = (SocketCount2 > 2 and jId32 == "0" and DefaultJewel[1]) or jId32
			end
			if iId3 then
				jId13 = (SocketCount3 > 0 and jId13 == "0" and DefaultJewel[1]) or jId13
				jId23 = (SocketCount3 > 1 and jId23 == "0" and DefaultJewel[1]) or jId23
				jId33 = (SocketCount3 > 2 and jId33 == "0" and DefaultJewel[1]) or jId33
			end
		elseif db.profile.Modifications.Jewels == L["Always default"] then
			jId11 = (SocketCount1 > 0 and DefaultJewel[1]) or "0"
			jId21 = (SocketCount1 > 1 and DefaultJewel[1]) or "0"
			jId31 = (SocketCount1 > 2 and DefaultJewel[1]) or "0"
			if iId2 then
				jId12 = (SocketCount2 > 0 and DefaultJewel[1]) or "0"
				jId22 = (SocketCount2 > 1 and DefaultJewel[1]) or "0"
				jId32 = (SocketCount2 > 2 and DefaultJewel[1]) or "0"
			end
			if iId3 then
				jId13 = (SocketCount2 > 0 and DefaultJewel[1]) or "0"
				jId23 = (SocketCount2 > 1 and DefaultJewel[1]) or "0"
				jId33 = (SocketCount2 > 2 and DefaultJewel[1]) or "0"
			end
		elseif db.profile.Modifications.Jewels == L["Ignore"] then
			jId11, jId21, jId31 = "0", "0", "0"
			if iId2 then jId12, jId22, jId32 = "0", "0", "0" end
			if iId3 then jId13, jId23, jId33 = "0", "0", "0" end
		end
	end
	
	ItemString1 = GetItemString(iId1, eId1, jId11, jId21, jId31, jId41, sId1, uId1)
	if iId2 then ItemString2 = GetItemString(iId2, eId2, jId12, jId22, jId32, jId42, sId2, uId2) end
	if iId3 then ItemString3 = GetItemString(iId3, eId3, jId13, jId23, jId33, jId43, sId3, uId3) end
	
	--Print(self, ItemString1, ItemString2, ItemString3)
	
	for SetName, Set in pairs(SetByName) do
		local SetPoints1, SetPoints2, SetPoints3
		
		if Set.Enabled then
			SetPoints1 = round(Set:GetItemValue(ItemString1) or 0, 1)
			if ItemString2 then SetPoints2 = round(Set:GetItemValue(ItemString2) or 0, 1) end
			if ItemString3 then SetPoints3 = round(Set:GetItemValue(ItemString3) or 0, 1) end
			
			if (SetPoints1 > 0 
			 and not (SetPoints2 and SetPoints1 == SetPoints2)
			 and not (SetPoints3 and SetPoints1 == SetPoints3))
			 or (SetPoints2 and SetPoints1 ~= SetPoints2)
			 or (SetPoints3 and SetPoints1 ~= SetPoints3) then
				
				local line = (SetPoints2 or SetPoints3) and Compare(db.profile.Tooltip.Compare, 
				db.profile.Tooltip.ShowPoints, db.profile.Tooltip.SwapColors, db.profile.Tooltip.SwapComparison, SetPoints1, SetPoints2, SetPoints3) or SetPoints1
				
				if db.profile.Tooltip.RightSide then
					text = ((text and (text ~= "") and (text .. "\n")) or "") .. SetName
					text2 = ((text2 and (text2 ~= "") and (text2 .. "\n")) or "") .. line
				else
					text = ((text and (text ~= "") and (text .. "\n")) or "") .. SetName .. ": " .. line
				end
			end
		end
	end
	
	return text or "", text2 or ""
end

function RegisterNewJewel(JewelId)
	local ItemLink = GetItemLink(6948, 0, JewelId)
	local GemName, GemItemLink = GetItemGem(ItemLink, 1)
	if GemName then
		local GemItemString = ItemLinkToItemString(GemItemLink)
		local _, GemItemId = strsplit(":", GemItemString)
		GemItemId = tonumber(GemItemId) 
		local _, _, GemQuality = GetItemInfo(GemItemId)
		
		--g:SetHyperlink("item:21877:0:" .. cur .. ":0:0:0:0:0")
		--local effect = g:GetLine(2)
		
		Gratuity:SetHyperlink("item:"..GemItemId..":0:0:0:0:0:0:0")
		local numlines = Gratuity:NumLines()
		local ColorString = string.lower(Gratuity:GetLine(numlines))
		local red, yellow, blue, meta
		if string.find(ColorString, "red") then red = true end
		if string.find(ColorString, "yellow") then yellow = true end
		if string.find(ColorString, "blue") then blue = true end
		if string.find(ColorString, "meta") then meta = true end 
		
		local GemIsUnique = not not Gratuity:Find("Unique")
		
		local Jewel = {GemName, tonumber(JewelId), GemItemId, GemQuality, GemIsUnique, (red == true), (yellow == true), IsBlue = (blue == true), IsMeta = (meta == true),}
		
		Print(self, "New Jewel:", GemItemLink)
		
		table.insert(db.profile.Jewels, Jewel)
		
		JewelByEnchantId[Jewel[1]] = Jewel
		JewelByName[Jewel[3]] = Jewel
		
		Options.args.Jewels.args[Jewel[3]] = {
			order	= 101,
			type	= "toggle",
			name	= Jewel[3],
			desc	= Jewel[3],
			get		= function() return db.profile.EnabledJewels[Jewel[3]] end,
			set		= function(v) db.profile.EnabledJewels[Jewel[3]] = v ClearCache() end,
		}
	end
end

local color = {"|cFFFFFFFF[", "|cFF1EFF00[", "|cFF0070DD[", "|cFFA335EE[", "|cFFFF8000["}
function PrintBestJewels(self, SetName, ItemLink)
	local Set = SetByName[SetName]
	if not (ItemLink and Set) then return end
		
	local ItemString = ItemLinkToItemString(ItemLink)
	local _, iId, eId = strsplit(":", ItemString)
	
	if Set.Enabled then
		local ItemLink = GetItemLink(iId, eId)
		Print(self, string.format(L["Best jewel combination for %s (%s Points), Set: %s is:"], ItemLink, round(Set:GetItemValue(iId, eId), 1), SetName))
		
		local BestJewelCombination = CalcBestJewels(Set, ItemString)
		local Jewel1, Jewel2, Jewel3 = BestJewelCombination.Jewel1, BestJewelCombination.Jewel2, BestJewelCombination.Jewel3
		
		local _, ItemId, EnchantId = strsplit(":", ItemString)
		local ItemLink = GetItemLink(ItemId, EnchantId, Jewel1[1], Jewel2[1], Jewel3[1])
		
		Print(self, string.format(L[" %s %s Points:"], ItemLink, round(BestJewelCombination.Points, 1)))
		
		if Jewel1 and Jewel1 ~= EmptySocket then
			local _, ItemLinkJ1 = GetItemInfo(Jewel1[2])
			Print(self, " 1. ", (ItemLinkJ1 or color[Jewel1.Quality] .. Jewel1[3] .. "]|r (*)"))
		end
		if Jewel2 and Jewel2 ~= EmptySocket then
			local _, ItemLinkJ2 = GetItemInfo(Jewel2[2])
			Print(self, " 2. ", (ItemLinkJ2 or color[Jewel2.Quality] .. Jewel2[3] .. "]|r (*)"))
		end
		if Jewel3 and Jewel3 ~= EmptySocket then
			local _, ItemLinkJ3 = GetItemInfo(Jewel3[2])
			Print(self, " 3. ", (ItemLinkJ3 or color[Jewel3.Quality] .. Jewel3[3] .. "]|r (*)"))
		end
	end
end

function CalcBestJewels(Set, ItemString)
	if not ItemString then return end
	local _, iId, eId, jId1, jId2, jId3, jId4, sId, uId = strsplit(":", ItemString)
	if not iId then iId = ItemString end
	
	local Sockets, RedSockets, YellowSockets, BlueSockets, MetaSocketPos = GetSocketInfo(iId)
	
	if not Sockets then return ItemString end
	
	-- 3 best jewels for every type are calculated
	local BestJewelRed, BestJewelYellow, BestJewelBlue, BestJewelMeta = {}, {}, {}, {}
	
	for JewelName, Jewel in pairs(JewelByName) do
		if db.profile.EnabledJewels[JewelName] then
			-- Cant scan jewel's ItemId directly, because its not sure if its in the clients item cache
			-- Workaround: Scan Hearthstone (6948) with inserted gem
			
			-- Used later to sort
			local JewelPoints = {
				Name = JewelName,
				Points = Set:GetItemValue(GetItemString(6948, 0, Jewel[1], 0, 0, 0, 0, 0)),
			}
			
			if Jewel[6] then
				table.insert(BestJewelRed, JewelPoints)
			elseif Jewel[7] then
				table.insert(BestJewelYellow, JewelPoints)
			elseif Jewel[8] then
				table.insert(BestJewelBlue, JewelPoints)
			elseif Jewel[9] and MetaSocketPos then
				table.insert(BestJewelMeta, JewelPoints)
			end
		end
	end
	
	assert(#BestJewelRed >= 3, "#BestJewelRed is "..#BestJewelRed)
	assert(#BestJewelYellow >= 3, "#BestJewelYellow is "..#BestJewelYellow)
	assert(#BestJewelBlue >= 3, "#BestJewelBlue is "..#BestJewelBlue)
	
	table.sort(BestJewelRed, function (v1, v2) return v1.Points > v2.Points end)
	table.sort(BestJewelYellow, function (v1, v2) return v1.Points > v2.Points end)
	table.sort(BestJewelBlue, function (v1, v2) return v1.Points > v2.Points end)
	table.sort(BestJewelMeta, function (v1, v2) return v1.Points > v2.Points end)
	
	-- The up to 9 Gems that are most likely to build the best combination
	local JewelList = {}
	
	if RedSockets or BestJewelRed[1].Points > BestJewelYellow[1].Points or BestJewelRed[1].Points > BestJewelBlue[1].Points then
		table.insert(JewelList, JewelByName[BestJewelRed[1][3]])
		if JewelByName[BestJewelRed[1].Name].IsUnique then
			table.insert(JewelList, JewelByName[BestJewelRed[2][3]])
			if JewelByName[BestJewelRed[2].Name].IsUnique then
				table.insert(JewelList, JewelByName[BestJewelRed[3][3]])
			end
		end
	end
	
	if YellowSockets or BestJewelYellow[1].Points > BestJewelRed[1].Points or BestJewelYellow[1].Points > BestJewelBlue[1].Points then
		table.insert(JewelList, JewelByName[BestJewelYellow[1].Name])
		if JewelByName[BestJewelYellow[1].Name].IsUnique then
			table.insert(JewelList, JewelByName[BestJewelYellow[2].Name])
			if JewelByName[BestJewelYellow[2].Name].IsUnique then
				table.insert(JewelList, JewelByName[BestJewelYellow[3].Name])
			end
		end
	end
	
	if BlueSockets or BestJewelBlue[1].Points > BestJewelYellow[1].Points or BestJewelBlue[1].Points > BestJewelRed[1].Points then
		table.insert(JewelList, JewelByName[BestJewelBlue[1].Name])
		if JewelByName[BestJewelBlue[1].Name].IsUnique then
			table.insert(JewelList, JewelByName[BestJewelBlue[2].Name])
			if JewelByName[BestJewelBlue[2].Name].IsUnique then
				table.insert(JewelList, JewelByName[BestJewelBlue[3].Name])
			end
		end
	end
	
	if not DummyJewelTable then DummyJewelTable = {EmptySocket} end
	
	local ScanTable1, ScanTable2, ScanTable3
	
	if MetaSocketPos > 0 then
		if not MetaGemTable then MetaGemTable = {} end
		MetaGemTable[1] = JewelByName[BestJewelMeta[1].Name]
	end
	
	ScanTable3 = (MetaSocketPos == 3 and MetaGemTable) or (#Sockets >= 3 and JewelList) or DummyJewelTable
	ScanTable2 = (MetaSocketPos == 2 and MetaGemTable) or (#Sockets >= 2 and JewelList) or DummyJewelTable
	ScanTable1 = (MetaSocketPos == 1 and MetaGemTable) or (JewelList)
	
	local BestJewelCombination = {
		Points = -1,
	}
	
	for _, Jewel1 in pairs(ScanTable1) do
		for _, Jewel2 in pairs(ScanTable2) do
			for _, Jewel3 in pairs(ScanTable3) do
				if Jewel1.IsUnique and (Jewel1==Jewel2 or Jewel1==Jewel3)
				 or Jewel2.IsUnique and (Jewel2==Jewel3) then break end
				local ItemString = GetItemString(iId, eId, Jewel1[1], Jewel2[1], Jewel3[1], jId4, sId, uId)
				local Points = Set:GetItemValue(ItemString)
				
				if Points > BestJewelCombination.Points then
					BestJewelCombination.Points = Points
					BestJewelCombination.Jewel1 = Jewel1
					BestJewelCombination.Jewel2 = Jewel2
					BestJewelCombination.Jewel3 = Jewel3
				end
			end
		end
	end
	
	return BestJewelCombination
end

-- Item Utility Functions
function GetSocketInfo(ItemId)
	Gratuity:SetHyperlink("item:" .. ItemId .. ":0:0:0:0:0:0:0")
	
	local NumLines = Gratuity:NumLines()
	
	if NumLines < 5 then
		return
	end
	
	local RedSockets, YellowSockets, BlueSockets, MetaSocketPos = 0, 0, 0, 0
	
	local Sockets = {}
	
	for curLine = 4, NumLines do
		local line = Gratuity:GetLine(curLine)
		if line == L["Red Socket"] then
			table.insert(Sockets, "red")
			RedSockets = RedSockets + 1
		elseif line == L["Yellow Socket"] then
			table.insert(Sockets, "yellow")
			YellowSockets = YellowSockets + 1
		elseif line == L["Blue Socket"] then
			table.insert(Sockets, "blue")
			BlueSockets = BlueSockets + 1
		elseif line == L["Meta Socket"] then
			table.insert(Sockets, "meta")
			MetaSocketPos = #Sockets
		end
		if #Sockets >= 3 then break end
	end
	
	if #Sockets == 0 then return end
	return Sockets, RedSockets, YellowSockets, BlueSockets, MetaSocketPos
end

function GetSocketCount(itemId)
	if SocketAmountCache[itemId] then
		return SocketAmountCache[itemId]
	else
		Gratuity:SetHyperlink("item:" .. itemId .. ":0:0:0:0:0:0:0")
		
		local NumLines = Gratuity:NumLines()
		
		if NumLines < 5 then
			return 0
		end
		
		local sockets = 0
		for curLine = 4, NumLines do
			local line = Gratuity:GetLine(curLine)
			if line == L["Red Socket"] or line == L["Yellow Socket"]
			 or line == L["Blue Socket"] or line == L["Meta Socket"] then
				sockets = sockets + 1
			end
			if sockets >= 3 then break end
		end
		
		SocketAmountCache[itemId] = sockets
		return sockets
	end
end

--[[
#############################
--    Utility functions
--############################
--]]

function round(num, idp)
	local mult = 10^(idp or 0)
	return math.floor(num * mult + 0.5) / mult
end

do
	local cBetter = "|cFF00FF00"
	local cWorse  = "|cFFFF0000"
	local cEqual = "|cFFFFFF00"
	
	local function Color(reverse, b, v)
		if reverse then
			if (b > v) then
				return cWorse
			elseif (b < v) then
				return cBetter
			else
				return cEqual
			end
		else
			if (b > v) then
				return cBetter
			elseif (b < v) then
				return cWorse
			else
				return cEqual
			end
		end
	end
	
	local function FormatNumber(n)
		if n > 0 then
			return "+" .. round(n, 1)
		else
			return round(n, 1) .. ""
		end
	end
	
	local CompModes = {
		--["Don't Compare"] = nil,
		["Absolute"] = function(b, v)
			return v
		end,
		["Delta"] = function(b, v)
			return FormatNumber(b-v)
		end,
		["Percent"] = function(b, v)
			return (b > 0 and (FormatNumber(v/b*100-100)).."%") or "+"
		end
	}
	
	function Compare(CompMode, ShowBase, ReverseColor, SwapComparison, ...)
		local nargs = select("#", ...)
		local b = select(1, ...)
		
		local line = {}
		
		if ShowBase then
			--if nargs == 2 then table.insert(line, GetColor(select(2, ...), b)) end
			table.insert(line, b)
			--if nargs == 2 then table.insert(line, "|r") end
		end
		
		local i = 2
		local v = select(2, ...)
		while v do
			if CompModes[CompMode] then
				if not v then
					Print(self, ...)
				end
				table.insert(line, Color(ReverseColor, b, v))
				table.insert(line, " (")
				if SwapComparison then
					table.insert(line, CompModes[CompMode](v, b))
				else
					table.insert(line, CompModes[CompMode](b, v))
				end
				table.insert(line, ")|r")
			end
			i = i + 1
			v = select(i, ...)
		end
		
		return table.concat(line)
	end
end

do
	local slot1 = {
		INVTYPE_AMMO			= 0,
		INVTYPE_GUNPROJECTILE	= 0,
		INVTYPE_BOWPROJECTILE	= 0,
		INVTYPE_HEAD			= 1,
		INVTYPE_NECK			= 2,
		INVTYPE_SHOULDER		= 3,
		INVTYPE_BODY			= 4,
		INVTYPE_CHEST			= 5,
		INVTYPE_ROBE			= 5,
		INVTYPE_WAIST			= 6,
		INVTYPE_LEGS			= 7,
		INVTYPE_FEET			= 8,
		INVTYPE_WRIST			= 9,
		INVTYPE_HAND			= 10,
		INVTYPE_FINGER			= 11,
		INVTYPE_TRINKET			= 13,
		INVTYPE_CLOAK			= 15,
		INVTYPE_WEAPON			= 16,
		INVTYPE_2HWEAPON		= 16,
		INVTYPE_WEAPONMAINHAND	= 16,
		INVTYPE_WEAPONOFFHAND	= 17,
		INVTYPE_SHIELD			= 17,
		INVTYPE_HOLDABLE		= 17,
		INVTYPE_RANGED			= 18,
		INVTYPE_RANGEDRIGHT		= 18,
		INVTYPE_RELIC			= 18,
		INVTYPE_GUN				= 18,
		INVTYPE_CROSSBOW		= 18,
		INVTYPE_WAND			= 18,
		INVTYPE_THROWN			= 18,
		INVTYPE_TABARD			= 19,
	}

	local slot2 = {
		INVTYPE_FINGER		= 12,
		INVTYPE_TRINKET		= 14,
		INVTYPE_WEAPON		= 17,
		INVTYPE_2HWEAPON	= 17,
	}

	function GetInventorySlot(ItemLink)
		local _, _, _, _, _, _, _, _, ItemEquipLoc = GetItemInfo(ItemLink)
		if not ItemEquipLoc then return end
		
		local s1, s2 = slot1[ItemEquipLoc], slot2[ItemEquipLoc]
		
		return s1, s2
	end
end

function ItemLinkToItemString(ItemLink)
	if not ItemLink then return end
	local il, _,ItemString = string.find(ItemLink, "^|%x+|H(.+)|h%[.+%]")
	
	return il and ItemString or ItemLink
end

function GetItemString(itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId)
	return "item:"..itemId..":"..(enchantId or 0)..":"..(jewelId1 or 0)..":"..(jewelId2 or 0)..":"..(jewelId3 or 0)..":"..(jewelId4 or 0)..":"..(suffixId or 0)..":"..(uniqueId or 0)
end

function GetItemLink(itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId)
	local itemString = GetItemString(itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId)
	local _, itemLink = GetItemInfo(itemString)
	return itemLink
end

function ItemStringToItemLink(itemString)
	local _, itemLink = GetItemInfo(itemString)
	return itemLink
end
